# Generated by Django 3.2.9 on 2021-12-04 02:51

import ckeditor.fields
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('email_signals', '0003_auto_20211204_0211'),
    ]

    operations = [
        migrations.AlterField(
            model_name='signal',
            name='html_email',
            field=ckeditor.fields.RichTextField(blank=True, null=True, verbose_name='HTML email'),
        ),
    ]




from django.test import SimpleTestCase
from django.core import mail
from .testcase import setup_settings
from .. import emailer


class TestSendMail(SimpleTestCase):
    """Unittests for the `send_mail` function in the `emailer` module."""

    def test_send_mail_no_template(self):
        """Tests that the function is able to send an email without a
        template."""
        emailer.send_mail(
            subject='Test Subject',
            plain_message='Test Message',
            html_message='<p>Test Message</p>',
            recipient_list=['test@test.com'],
            from_email='test@test.com'
        )
        self.assertEqual(len(mail.outbox), 1)

    def test_send_mail_with_template(self):
        """Tests that the function is able to send an email with a template."""

        setup_settings()

        emailer.send_mail(
            subject='Test Subject',
            plain_message='Test Message',
            html_message='<p>Test Message</p>',
            recipient_list=['test@test.com'],
            from_email='test@test.com',
            template='email_signals/tests/test_emailer.html',
        )
        self.assertEqual(len(mail.outbox), 1)



from django import forms
from django.template.loader import get_template
from django.template import TemplateDoesNotExist
from django.contrib.contenttypes.models import ContentType
from . import models
from .registry import registered_content_types
from .utils import get_param_from_obj
from .constraint_checker import comparison_requires_2_params


class SignalAdminForm(forms.ModelForm):

    class Meta:
        model = models.Signal
        fields = '__all__'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Limit the content type choices to the registered content types
        self.fields['content_type'].queryset = registered_content_types()

    def clean(self):
        cleaned_data = super().clean()
        if self.is_valid():
            self._clean_mailing_list()
        return cleaned_data

    def _clean_mailing_list(self):
        """The `mailing_list` fill contains a string which corresponds to a
        function that should exist in the model instance. Check that the
        function does exists.
        """
        mailing_list = self.cleaned_data['mailing_list']
        content_type = self.cleaned_data['content_type']

        if not hasattr(content_type.model_class(), mailing_list):
            raise forms.ValidationError(
                f"The model does not have a function called {mailing_list}"
            )
        return mailing_list

    def clean_template(self):
        """Check that the template exists."""
        template = self.cleaned_data['template']
        if not template:
            return template
        try:
            get_template(template)
        except TemplateDoesNotExist:
            raise forms.ValidationError(
                f"The template {template} does not exist"
            )
        return template


class SignalConstraintAdminForm(forms.ModelForm):

    class Meta:
        models = models.SignalConstraint
        fields = '__all__'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Set autocomplete off for all fields.
        for field in self.fields:
            self.fields[field].widget.attrs['autocomplete'] = 'off'

    def clean(self):
        cleaned_data = super().clean()
        if self.is_valid():
            self._clean_comparison()
        return cleaned_data

    def clean_param_1(self):
        """Validate `param_2` can be found in either the signal `kwargs` or the
        model instance.
        """
        param_1 = self.cleaned_data['param_1']

        valid, _ = get_param_from_obj(param_1, self.instance)
        if not valid:
            raise forms.ValidationError(
                f"The model does not have a parameter called {param_1}"
            )
        return param_1

    def _clean_comparison(self):
        """Validate the comparison is valid for the given parameters.
        Depending on what the `comparison` is, it would limit the value for
        `param_1` and `param_2`.

        Note: Due to the order of fields, we need to preprend the underscore
        in the function name and call it manually as `param_2` is cleaned
        after 'comparison'.
        """

        comparison = self.cleaned_data['comparison']
        param_2 = self.cleaned_data['param_2']

        if comparison_requires_2_params(comparison) and not param_2:
            raise forms.ValidationError(
                "This comparison requires a second parameter"
            )

        if not comparison_requires_2_params(comparison) and param_2:
            raise forms.ValidationError(
                "This comparison does not require a second parameter"
            )

        return comparison



from django.contrib import admin
from . import models, forms


class SignalConstraintInline(admin.TabularInline):
    form = forms.SignalConstraintAdminForm
    model = models.SignalConstraint
    extra = 0


@admin.register(models.Signal)
class SignalAdmin(admin.ModelAdmin):
    list_display = ('name', 'content_type', 'signal_type', 'constraints_count',
                    'active',)

    form = forms.SignalAdminForm
    change_form_template = 'email_signals/admin/email_signals/signal/change_form.html'  # noqa: E501
    inlines = [SignalConstraintInline]



import typing as _t
from django.db import models
from django.db.models import signals
from django.contrib.contenttypes.models import ContentType
from ckeditor.fields import RichTextField


class EmailSignalMixin:
    """A mixin to add a model to the email_signals app."""

    EMAIL_SIGNAL_MODEL = True

    def email_signal_recipients(self, method_name: str) -> _t.List[str]:
        """Return a list of email addresses to send the signal to.

        Args:
            method_name: The name of the method which when called will return
                a mailing list.

        Returns:
            A list of email addresses to send emails to.

        """
        if not hasattr(self, method_name):
            raise NotImplementedError(
                f'{self.__class__.__name__} has no method {method_name}'
            )
        return getattr(self, method_name)()


class Signal(models.Model):
    """Stores signals to be raised by the email_signals app."""

    class SignalTypeChoices(models.TextChoices):
        """Choices for the type of signal."""
        pre_save = 'pre_save', 'Pre Save'
        post_save = 'post_save', 'Post Save'
        pre_delete = 'pre_delete', 'Pre Delete'
        post_delete = 'post_delete', 'Post Delete'

    name = models.CharField(max_length=255)
    description = models.TextField(blank=True, null=True)
    content_type = models.ForeignKey(
        ContentType,
        on_delete=models.CASCADE,
        verbose_name='Model'
    )
    plain_message = models.TextField(
        blank=True,
        null=True,
        verbose_name='Plain text content'
    )
    html_message = RichTextField(
        blank=True,
        null=True,
        verbose_name='HTML content'
    )
    subject = models.CharField(max_length=255)
    from_email = models.EmailField(
        null=True,
        blank=True,
        help_text='If not set, `settings.EMAIL_SIGNAL_DEFAULT_SENDER` \
            with be used.'
    )
    mailing_list = models.CharField(
        max_length=100,
        help_text="The mailing list to send the signal to. Will search for a \
            function with the same name in the model instance.",
    )
    template = models.CharField(
        max_length=100,
        blank=True,
        null=True,
        help_text="Custom template to use for the email. (Paths relative to \
            `settings.TEMPLATES[i]['DIRS']`)"
    )
    signal_type = models.CharField(
        max_length=20,
        choices=SignalTypeChoices.choices
    )
    active = models.BooleanField(default=True)

    def __str__(self) -> str:
        return f'({self.signal_type}): {self.name}'

    def get_signal_type(self) -> signals.ModelSignal:
        """Return the signal type."""
        return getattr(signals, self.signal_type)

    @classmethod
    def get_choice_from_signal(cls, signal: signals.ModelSignal) -> str:
        """Return the signal type from the signal."""
        if signal == signals.pre_save:
            return cls.SignalTypeChoices.pre_save
        elif signal == signals.post_save:
            return cls.SignalTypeChoices.post_save
        elif signal == signals.pre_delete:
            return cls.SignalTypeChoices.pre_delete
        elif signal == signals.post_delete:
            return cls.SignalTypeChoices.post_delete
        else:
            raise ValueError(f'Unknown signal: {signal}')

    @classmethod
    def get_for_model_and_signal(
        cls,
        instance: 'Signal',
        signal: signals.ModelSignal
    ) -> models.QuerySet['Signal']:
        """Return the signal for the given model and signal type."""
        return cls.objects.filter(
            content_type=ContentType.objects.get_for_model(instance),
            signal_type=cls.get_choice_from_signal(signal),
            active=True,
        )

    @property
    def constraints_count(self) -> int:
        """Return the number of constraints."""
        return self.constraints.count()

    @property
    def model(self) -> models.base.ModelBase:
        """Return the model of the signal."""
        return self.content_type.model_class()

    def is_pre_save(self) -> bool:
        """Return `True` if the signal is a pre save signal."""
        return self.signal_type == self.SignalTypeChoices.pre_save

    def is_post_save(self) -> bool:
        """Return `True` if the signal is a post save signal."""
        return self.signal_type == self.SignalTypeChoices.post_save

    def is_pre_delete(self) -> bool:
        """Return `True` if the signal is a pre delete signal."""
        return self.signal_type == self.SignalTypeChoices.pre_delete

    def is_post_delete(self) -> bool:
        """Return `True` if the signal is a post delete signal."""
        return self.signal_type == self.SignalTypeChoices.post_delete


class SignalConstraint(models.Model):
    """Stores the constraints for a signal."""

    COMPARISON_CHOICES = (
        ('exact', 'Is Equal To'),
        ('iexact', 'Is Equal To (case insensitive)'),
        ('contains', 'Contains'),
        ('icontains', 'Contains (case insensitive)'),
        ('gt', 'Greater Than'),
        ('gte', 'Greater Than or Equal To'),
        ('lt', 'Less Than'),
        ('lte', 'Less Than or Equal To'),
        ('startswith', 'Starts With'),
        ('istartswith', 'Starts With (case insensitive)'),
        ('endswith', 'Ends With'),
        ('iendswith', 'Ends With (case insensitive)'),
        ('regex', 'Matches Regular Expression'),
        ('iregex', 'Matches Regular Expression (case insensitive)'),
        ('isnull', 'Is Null'),
        ('isnotnull', 'Is Not Null'),
        ('istrue', 'Is True'),
        ('isfalse', 'Is False'),
    )

    signal = models.ForeignKey(
        Signal,
        on_delete=models.CASCADE,
        related_name='constraints'
    )
    param_1 = models.CharField(
        max_length=255,
        verbose_name='Parameter 1',
        help_text='Will be searched in the instance and signal kwargs recursively. Use "." to show a layer in each attribute.'  # noqa
    )
    comparison = models.CharField(
        max_length=20,
        choices=COMPARISON_CHOICES
    )
    param_2 = models.CharField(
        max_length=255,
        blank=True,
        null=True,
        verbose_name='Parameter 2',
        help_text='Will be searched in the instance and signal kwargs recursively. Use "." to show a layer in each attribute. Also supports primitive values.'
    )

    def __str__(self) -> str:
        return f'{self.signal.name} - {self.comparison} - {self.param_1}'



import typing as _t
from django.contrib.contenttypes.models import ContentType
from .testcase import EmailSignalTestCase
from ..forms import SignalAdminForm, SignalConstraintAdminForm
from ..registry import add_to_registry


class TestSignalAdminForm(EmailSignalTestCase):
    """Unitests for the `SignalAdminForm` class."""

    @classmethod
    def setUpClass(cls):
        """Creates a table for the model."""
        super().setUpClass()
        add_to_registry(cls.Customer)

    def sample_form(
        self,
        override_form_data: _t.Optional[dict] = None
    ) -> SignalAdminForm:
        form_data = {
            "name": "Test Signal",
            'description': 'Test description',
            'content_type': ContentType.objects.get_for_model(
                self.customer_rec
            ).pk,
            'plain_message': 'Test plain message',
            'html_message': 'Test html message',
            'subject': 'Test subject',
            'from_email': 'test@email.com',
            'mailing_list': 'my_mailing_list',
            'template': None,
            'signal_type': 'pre_save'
        }
        form_data.update(override_form_data or {})
        form = SignalAdminForm(data=form_data)
        form.is_valid()
        return form

    def test_valid_form(self):
        """Basic test to check that a valid form is accepted."""
        form = self.sample_form()
        self.assertTrue(form.is_valid(), form.data)

    def test_invalid_mailing_list(self):
        """Test form where an invalid `mailing_list` is provided."""
        form = self.sample_form({'mailing_list': 'invalid'})
        self.assertFalse(form.is_valid())

    def test_invalid_template(self):
        """Test form where an invalid `template` is provided."""
        form = self.sample_form({'template': 'invalid'})
        self.assertFalse(form.is_valid())

    def test_valid_template(self):
        """Test form where a valid `template` is provided."""
        form = self.sample_form({
            'template':
            'email_signals/tests/test_emailer.html'
        })
        self.assertTrue(form.is_valid(), form.data)



"""Contains utility functions for the email_signals package."""

import typing as _t
from django.db.models.base import ModelBase
from django.db.models.fields.related_descriptors import ManyToManyDescriptor


def convert_to_primitive(param: str) -> _t.Any:
    """Converts the `param` to a primitive type is possible.

    Returns:
        bool: True if the `param` was converted to a primitive type.
        _t.Any: The converted value.
    """
    if param.lower() == 'true':
        return True
    if param.lower() == 'false':
        return False
    if param.lower() in ('none', 'null'):
        return None
    if '.' in param:
        try:
            return float(param)
        except ValueError:
            pass
    try:
        return int(param)
    except ValueError:
        pass
    return param


def get_param_from_obj(
    param: str,
    searchable: object,
    seperator: str = '.'
) -> (bool, _t.Any):
    """Search for the `param` in the `searchable` object. Iteratively
    search through the object's attributes until the `param` is found.

    Note: Though technically possible, this function will not call a function
    if a function is found as this could be a security risk.

    Args:
        param: The param to get the value for.
        searchable: The object to search for the param.
        seperator: The seperator to split the param by.

    Returns:
        bool: True if the `param` was found in the `self.instance` object.
        _t.Any: The value of the `param`.
    """
    param_parts = param.split(seperator)
    current_object = searchable

    while param_parts:
        param_part = param_parts.pop(0)

        # Need to search in a dictionary separately as `hasattr` is not
        # supported for dictionaries of python 3.8 and below.
        if isinstance(current_object, dict):
            if param_part in current_object:
                current_object = current_object[param_part]
                continue
            else:
                return False, None

        try:
            if hasattr(current_object, param_part):
                current_object = getattr(current_object, param_part)
                continue
        except Exception:
            return False, None

        # Test if the current object is iterable.
        if param_part.isdigit():
            try:
                iter_object = list(iter(current_object))
                if len(iter_object) > int(param_part):
                    current_object = iter_object[int(param_part)]
                    continue
            except TypeError:
                return False, None

        if param_part in ('self', 'instance'):
            return True, current_object

        return False, None

    return True, current_object


def get_model_attr_names(model_class: ModelBase, seen_attrs=None) -> dict:
    """Recursively, get all the attribute names from a model class.

    Note: This function will not traverse into the `ManyToManyDescriptor`. As
    we keep a set of seen attributes, by traversing into the
    `ManyToManyDescriptor` we end up adding an attribute inside a deeper level
    of the model class which should be in a higher level.

    Args:
        model_class: The model class to get the attribute names from.
        seen_attrs: A set of objects IDs that have already been seen.

    Returns:
        dict: The attribute names of the model attributes and their subsequent
            children names.
    """
    attr_names = {}
    seen_attrs = seen_attrs or set()
    for attr_name in dir(model_class):

        # Skip private attributes.
        if attr_name.startswith('_'):
            continue

        attr = getattr(model_class, attr_name)

        # A level of safety to prevent infinite recursion.
        if id(attr) in seen_attrs:
            continue
        seen_attrs.add(id(attr))

        # We don't call functions for the user to prevent security risks.
        # For that reason, we won't get the attribute names of functions.
        if callable(attr):
            continue
        attr_names[attr_name] = {}

        # If the attribute is a foreign key, get the attribute names of the
        # related object.
        if (hasattr(attr, 'field') and attr.field.is_relation
                and not isinstance(attr, ManyToManyDescriptor)):
            related_model = attr.field.related_model
            if related_model != attr.field.model:
                attr_names[attr_name] = get_model_attr_names(
                    related_model,
                    seen_attrs
                )

    return attr_names



# Generated by Django 3.2.9 on 2021-12-04 12:58

import ckeditor.fields
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('email_signals', '0004_alter_signal_html_email'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='signal',
            name='html_email',
        ),
        migrations.RemoveField(
            model_name='signal',
            name='plain_text_email',
        ),
        migrations.AddField(
            model_name='signal',
            name='html_message',
            field=ckeditor.fields.RichTextField(blank=True, null=True, verbose_name='HTML content'),
        ),
        migrations.AddField(
            model_name='signal',
            name='plain_message',
            field=models.TextField(blank=True, null=True, verbose_name='Plain text content'),
        ),
        migrations.AlterField(
            model_name='signal',
            name='content_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype', verbose_name='Model'),
        ),
        migrations.AlterField(
            model_name='signal',
            name='from_email',
            field=models.EmailField(blank=True, help_text='If not set, `settings.EMAIL_SIGNAL_DEFAULT_SENDER`             with be used.', max_length=254, null=True),
        ),
    ]



from django.core import mail
from django.db.models import signals as django_signals
from .testcase import EmailSignalTestCase
from .. import signals, models


class TestSignals(EmailSignalTestCase):
    """Test signals."""

    def test_signal_callback_with_no_constraints(self):
        """Test signal callback where there are no additional constraints
        applied to a signal.
        """
        self.create_signal(self.customer_rec)
        signals.signal_callback(self.customer_rec, django_signals.pre_save)
        self.assertEqual(len(mail.outbox), 1)

    def test_signal_callback_with_template(self):
        """Test signal callback where a template is applied to a signal."""
        self.create_signal(
            self.customer_rec,
            template='email_signals/tests/test_emailer.html',
        )
        signals.signal_callback(
            self.customer_rec,
            django_signals.pre_save
        )
        self.assertEqual(len(mail.outbox), 1)

    def test_signal_failing_constraints(self):
        """Test signal callback where the model instance fails the constraint
        checker. The expected outcome is that the email should not be sent.
        """
        signal = self.create_signal(
            self.customer_rec,
            template='email_signals/tests/test_emailer.html',
        )
        # This constraint will fail as it will attempt to check if
        # `instance.id` is equal to '-1'
        models.SignalConstraint.objects.create(
            signal=signal,
            param_1='id',
            param_2='-1',
            comparison='exact'
        )
        signals.signal_callback(
            self.customer_rec,
            django_signals.pre_save
        )
        self.assertEqual(len(mail.outbox), 0)

    def test_signal_passing_constraints(self):
        """Test signal callback where the model instance passes the constraint
        checker. The expected outcome is that the email should be sent.
        """
        signal = self.create_signal(
            self.customer_rec,
            template='email_signals/tests/test_emailer.html',
        )
        # This constraint will pass as it will attempt to check if
        # `instance.id` is greater than or equal to `1`
        models.SignalConstraint.objects.create(
            signal=signal,
            param_1='id',
            param_2='1',
            comparison='gte'
        )
        signals.signal_callback(
            self.customer_rec,
            django_signals.pre_save
        )
        self.assertEqual(len(mail.outbox), 1)

    def test_pre_save_signal(self):
        """Test the `pre_save` signal is triggered correctly."""
        self.setup_signals()
        self.create_signal(self.customer_rec).save()
        self.Customer.create_record()
        # Runs twice as the create record methods saves it after creation.
        self.assertEqual(len(mail.outbox), 2)

    def test_post_save_signal(self):
        """Test the `post_save` signal is triggered correctly."""
        self.setup_signals()
        self.create_signal(
            self.customer_rec,
            signal_type=models.Signal.SignalTypeChoices.post_save
        ).save()

        self.Customer.create_record()
        # Runs twice as the create record methods saves it after creation.
        self.assertEqual(len(mail.outbox), 2)

    def test_pre_delete_signal(self):
        """Test the `pre_delete` signal is triggered correctly."""
        self.setup_signals()
        self.create_signal(
            self.customer_rec,
            signal_type=models.Signal.SignalTypeChoices.pre_delete
        ).save()

        rec = self.Customer.create_record()
        rec.delete()
        self.assertEqual(len(mail.outbox), 1)

    def test_post_delete_signal(self):
        """Test the `post_delete` signal is triggered correctly."""
        self.setup_signals()
        self.create_signal(
            model_instance=self.customer_rec,
            signal_type=models.Signal.SignalTypeChoices.post_delete
        ).save()

        rec = self.Customer.create_record()
        rec.delete()
        self.assertEqual(len(mail.outbox), 1)

    def test_on_created_only(self):
        """Test that the signal is only called when a record is created."""
        self.setup_signals()
        signal = self.create_signal(
            self.customer_rec,
            signal_type=models.Signal.SignalTypeChoices.post_save
        )
        signal.save()
        models.SignalConstraint.objects.create(
            signal=signal,
            param_1='created',
            comparison='istrue'
        ).save()
        record = self.Customer.create_record()
        record.save()

        self.assertEqual(len(mail.outbox), 1)



from django.db.models import signals
from .testcase import EmailSignalTestCase
from .. import models


class TestEmailSignalMixin(EmailSignalTestCase):
    """Unittests for the `EmaiLSignalMixin`."""

    def test_email_signal_recipients(self):
        """Test the `email_signal_recipients` method."""

        self.assertIsInstance(
            self.customer_rec.email_signal_recipients('my_mailing_list'),
            list
        )

    def test_email_signal_recipients_invalid(self):
        """Test the `email_signal_recipients` method with an invalid function
        name. It should raise an `NotImplementedError`.
        """

        with self.assertRaises(NotImplementedError):
            self.customer_rec.email_signal_recipients('invalid_function')


class TestSignal(EmailSignalTestCase):
    """Unittests for the `Signal` model."""

    def test_str(self):
        """Test the `__str__` method."""
        self.assertIsInstance(str(self.create_signal(self.customer_rec)), str)

    def test_get_signal_type(self):
        """Test the `get_signal_type` method."""
        self.assertIsInstance(
            self.create_signal(self.customer_rec).get_signal_type(),
            signals.ModelSignal
        )

    def test_get_choice_from_signal(self):
        """Test the `get_choice_from_signal` method returns the correct
        signal type choice.
        """
        choices = models.Signal.SignalTypeChoices
        self.assertEqual(
            models.Signal.get_choice_from_signal(signals.pre_save),
            choices.pre_save
        )
        self.assertEqual(
            models.Signal.get_choice_from_signal(signals.post_save),
            choices.post_save
        )
        self.assertEqual(
            models.Signal.get_choice_from_signal(signals.pre_delete),
            choices.pre_delete
        )
        self.assertEqual(
            models.Signal.get_choice_from_signal(signals.post_delete),
            choices.post_delete
        )

    def test_get_choice_from_signal_invalid(self):
        """Test the `get_choice_from_signal` method with an invalid signal
        type. It should raise an `ValueError`.
        """
        with self.assertRaises(ValueError):
            models.Signal.get_choice_from_signal('invalid_signal')

    def test_get_for_model_and_signal(self):
        """Test the `get_for_model_and_signal` method. Given the model and
        signal, it should return a queryset of `Signal`.
        """
        qs = models.Signal.get_for_model_and_signal(
            self.customer_rec,
            signals.pre_save
        )

        for q in qs:
            self.assertEqual(q.model, self.customer_rec)
            self.assertEqual(q.signal, signals.pre_save)

    def test_constraints_count(self):
        """Test the `constraints_count` property. It should return the number
        of constraints.
        """
        signal = self.create_signal(self.customer_rec)
        self.assertEqual(signal.constraints_count, 0)

        models.SignalConstraint.objects.create(
            signal=signal,
            param_1='self',
            comparison='istrue'
        ).save()
        models.SignalConstraint.objects.create(
            signal=signal,
            param_1='self',
            comparison='gt',
            param_2=10
        ).save()

        self.assertEqual(signal.constraints_count, 2)

    def test_model(self):
        """Test the `model` property."""
        self.assertEqual(
            self.create_signal(self.customer_rec).model,
            self.customer_rec.__class__
        )

    def test_is_pre_save(self):
        """Test the `is_pre_save` property."""
        signal = self.create_signal(self.customer_rec)
        self.assertTrue(signal.is_pre_save())
        signal.signal_type = models.Signal.SignalTypeChoices.post_save
        self.assertFalse(signal.is_pre_save())

    def test_is_post_save(self):
        """Test the `is_post_save` property."""
        signal = self.create_signal(self.customer_rec)
        self.assertFalse(signal.is_post_save())
        signal.signal_type = models.Signal.SignalTypeChoices.post_save
        self.assertTrue(signal.is_post_save())

    def test_is_pre_delete(self):
        """Test the `is_pre_delete` property."""
        signal = self.create_signal(self.customer_rec)
        self.assertFalse(signal.is_pre_delete())
        signal.signal_type = models.Signal.SignalTypeChoices.pre_delete
        self.assertTrue(signal.is_pre_delete())

    def test_is_post_delete(self):
        """Test the `is_post_delete` property."""
        signal = self.create_signal(self.customer_rec)
        self.assertFalse(signal.is_post_delete())
        signal.signal_type = models.Signal.SignalTypeChoices.post_delete
        self.assertTrue(signal.is_post_delete())


class TestSignalConstraint(EmailSignalTestCase):
    """Unittests for the `SignalConstraint` model."""

    def test_str(self):
        signal = self.create_signal(self.customer_rec)
        signal_constraint = models.SignalConstraint.objects.create(
            signal=signal,
            param_1='self',
            comparison='istrue'
        )
        self.assertIsInstance(str(signal_constraint), str)



from django.db.models import signals
from django.contrib.contenttypes.models import ContentType
from ..models import Signal, SignalConstraint
from .testcase import EmailSignalTestCase
from email_signals.constraint_checker import (
    ConstraintChecker,
    comparison_requires_2_params
)


class TestConstraintChecker(EmailSignalTestCase):
    """Unittests for the `ConstraintChecker` class."""

    def test_init_no_constraints(self):
        """Test the `__init__` method where there are no constraints for a
        signal.
        """
        self.create_signal(self.customer_rec)
        constraint_checker = ConstraintChecker(self.customer_rec, {})
        self.assertEqual(list(constraint_checker.constraints), [])

    def test_init_with_constraints(self):
        """Test the `__init__` method where there are constraints for a
        signal.
        """
        signal_rec = self.create_signal(self.customer_rec)
        signal_constraint = SignalConstraint.objects.create(
            signal=signal_rec,
            param_1='a',
            comparison='eq',
            param_2='b',
        )
        signal_constraint.save()
        constraint_checker = ConstraintChecker(self.customer_rec, {})

        self.assertEqual(
            list(constraint_checker.constraints),
            [signal_constraint]
        )

    def test_check_constraint(self):
        """Test the `check_constraint` method."""
        self.assertTrue(ConstraintChecker.check_constraint('a', 'a', 'exact'))
        self.assertFalse(ConstraintChecker.check_constraint('a', 'b', 'exact'))

    def test_check_constraint_invalid_comparison(self):
        """Test the `check_constraint` method with an invalid comparison."""
        with self.assertRaises(ValueError):
            ConstraintChecker.check_constraint('a', 'b', 'abc')

    def test_get_params_bad_param_1(self):
        """Test the `get_param_1` method where the `param_1` is invalid. This
        should raise a `ValueError` as `param_1` is expected to exist in
        some level of the model instance or the signal kwargs.
        """
        signal = self.create_signal(self.customer_rec)
        constraint = SignalConstraint.objects.create(
            signal=signal,
            param_1='zzz',
            comparison='eq',
            param_2='b',
        )
        constraint.save()
        constaint_checker = ConstraintChecker(self.customer_rec, {})

        with self.assertRaises(ValueError):
            constaint_checker.get_param_1(constraint.param_1)

    def test_get_param_1_in_kwargs(self):
        """Test the `get_param_1` method where the `param_1` is in the
        signal kwargs.
        """
        signal = self.create_signal(self.customer_rec)
        constraint = SignalConstraint.objects.create(
            signal=signal,
            param_1='a.b',
            comparison='eq',
            param_2=None,
        )
        constraint.save()
        constaint_checker = ConstraintChecker(
            self.customer_rec,
            {'a': {'b': 1}}
        )

        self.assertEqual(constaint_checker.get_param_1(constraint.param_1), 1)

    def test_get_param_1_in_instance(self):
        """Test the `get_param_1` method where the `param_1` is in the
        model instance.
        """
        self.customer_rec.name = 'Test Name'
        self.customer_rec.save()

        signal = self.create_signal(self.customer_order_rec)
        constraint = SignalConstraint.objects.create(
            signal=signal,
            param_1='customer.name',
            comparison='eq',
            param_2=None,
        )
        constraint.save()
        constaint_checker = ConstraintChecker(
            self.customer_order_rec,
            {}
        )

        self.assertEqual(
            constaint_checker.get_param_1(constraint.param_1),
            'Test Name'
        )

    def test_get_param_2_in_kwargs(self):
        """Test the `get_param_2` method where the `param_2` is in the
        signal kwargs.
        """
        signal = self.create_signal(self.customer_rec)
        constraint = SignalConstraint.objects.create(
            signal=signal,
            param_1='a',
            comparison='eq',
            param_2='a.b',
        )
        constraint.save()
        constaint_checker = ConstraintChecker(
            self.customer_rec,
            {'a': {'b': 1}}
        )

        self.assertEqual(constaint_checker.get_param_1(constraint.param_2), 1)

    def test_get_param_2_in_instance(self):
        """Test the `get_param_2` method where the `param_2` is in the
        model instance.
        """
        self.customer_rec.name = 'Test Name'
        self.customer_rec.save()

        signal = self.create_signal(self.customer_order_rec)
        constraint = SignalConstraint.objects.create(
            signal=signal,
            param_1='a',
            comparison='eq',
            param_2='customer.name',
        )
        constraint.save()
        constaint_checker = ConstraintChecker(
            self.customer_order_rec,
            {}
        )

        self.assertEqual(
            constaint_checker.get_param_2(constraint.param_2),
            'Test Name'
        )

    def test_get_param_2_in_primitive(self):
        """Test the `get_param_2` method where the `param_2` can be converted
        into a primitive value.
        """
        signal = self.create_signal(self.customer_rec)
        constraint = SignalConstraint.objects.create(
            signal=signal,
            param_1='a',
            comparison='eq',
            param_2='b',
        )
        constraint.save()
        constaint_checker = ConstraintChecker(
            self.customer_rec,
            {}
        )

        self.assertEqual(
            constaint_checker.get_param_2(constraint.param_2),
            'b'
        )

    def test_get_params(self):
        """Test the `get_params` method."""
        signal = self.create_signal(self.customer_rec)
        constraint = SignalConstraint.objects.create(
            signal=signal,
            param_1='a',
            comparison='eq',
            param_2='1.1',
        )
        constraint.save()
        constaint_checker = ConstraintChecker(
            self.customer_rec,
            {'a': 1}
        )

        self.assertEqual(constaint_checker.get_params(constraint), (1, 1.1))

    def test_run_tests_passes(self):
        """Test the `run_tests` method where the constraints should pass."""
        signal = self.create_signal(self.customer_rec)
        constraint = SignalConstraint.objects.create(
            signal=signal,
            param_1='a',
            comparison='exact',
            param_2='1.1',
        )
        constraint.save()
        constaint_checker = ConstraintChecker(
            self.customer_rec,
            {'a': 1.1}
        )

        self.assertTrue(constaint_checker.run_tests())

    def test_run_tests_fails(self):
        """Test the `run_tests` method where the constraints should fail."""
        signal = self.create_signal(self.customer_rec)
        constraint = SignalConstraint.objects.create(
            signal=signal,
            param_1='a',
            comparison='exact',
            param_2='1.1',
        )
        constraint.save()
        constaint_checker = ConstraintChecker(
            self.customer_rec,
            {'a': 1}
        )

        self.assertFalse(constaint_checker.run_tests())

    def test_comparison_requires_2_params(self):
        """Test the `comparison_requires_2_params` method."""
        self.assertTrue(comparison_requires_2_params('exact'))
        self.assertFalse(comparison_requires_2_params('istrue'))



"""Dynamically creates signals for registered models."""

from functools import partial
from django.db.models import signals, Model
from .constraint_checker import ConstraintChecker
from . import models, emailer


def signal_callback(
    instance: Model,
    signal: signals.ModelSignal,
    **kwargs
) -> None:
    """Callback triggered by signals. This function will check if for a given
    model instance, certain constraints are met. If so, it will send an email.
    """
    
    model_signals = models.Signal.get_for_model_and_signal(instance, signal)
    for model_signal in model_signals:
        if not model_signal.active:
            continue
        if not ConstraintChecker(instance, kwargs).run_tests():
            continue

        # When the program reaches this point, the constraint checker has
        # passed.
        emailer.send_mail(
            subject=model_signal.subject,
            plain_message=model_signal.plain_message,
            html_message=model_signal.html_message,
            from_email=model_signal.from_email,
            recipient_list=instance.email_signal_recipients(
                model_signal.mailing_list
            ),
            template=model_signal.template,
            context={'instance': instance, 'signal_kwargs': kwargs},
        )


def setup():
    """Dynamically connections functions to signals for each model in the
    registry.
    """

    # Import needs to happen here for unittesting - otherwise the
    # changes `registered_models` to are not picked up by the tests.
    from .registry import registered_models

    # TODO: Add support for custom signals.

    signal_types = (
        signals.pre_save,
        signals.post_save,
        signals.pre_delete,
        signals.post_delete
    )

    signal_factory = []

    for model in registered_models.values():
        for signal_type in signal_types:
            signal_factory.append(partial(
                signal_type.connect,
                signal_callback,
                sender=model
            ))

    for function in signal_factory:
        function()



# Generated by Django 3.2.9 on 2021-12-04 13:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('email_signals', '0005_auto_20211204_1258'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='signal',
            name='to_emails_opt',
        ),
        migrations.AddField(
            model_name='signal',
            name='mailing_list',
            field=models.CharField(default=1, help_text='The mailing list to send the signal to. Will search for a             function with the same name in the model instance.', max_length=100),
            preserve_default=False,
        ),
        migrations.AlterField(
            model_name='signal',
            name='template',
            field=models.CharField(blank=True, help_text="Custom template to use for the email. (Paths relative to             `settings.TEMPLATES[i]['DIRS']`)", max_length=100, null=True),
        ),
    ]



# Generated by Django 3.2.9 on 2021-12-04 02:11

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        ('email_signals', '0002_auto_20211203_1357'),
    ]

    operations = [
        migrations.RenameField(
            model_name='signalconstraint',
            old_name='comparision',
            new_name='comparison',
        ),
        migrations.AlterField(
            model_name='signal',
            name='content_type',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.contenttype', verbose_name='Model (Table)'),
        ),
        migrations.AlterField(
            model_name='signal',
            name='html_email',
            field=models.TextField(blank=True, null=True, verbose_name='HTML email'),
        ),
        migrations.AlterField(
            model_name='signal',
            name='to_emails_opt',
            field=models.PositiveIntegerField(default=1, help_text='The choice of the user to send the email to. For each             integer `i`, the method `get_email_signal_emails_<i>` will be             called on the model instance. The method should return a list of             emails to send to.', verbose_name='Mailing list no'),
        ),
        migrations.AlterField(
            model_name='signalconstraint',
            name='param_1',
            field=models.CharField(help_text='Will be searched in the instance and signal kwargs recursively. Use "." to show a layer in each attribute.', max_length=255, verbose_name='Parameter 1'),
        ),
        migrations.AlterField(
            model_name='signalconstraint',
            name='param_2',
            field=models.CharField(blank=True, help_text='Will be searched in the instance and signal kwargs recursively. Use "." to show a layer in each attribute. Also supports primitive values.', max_length=255, null=True, verbose_name='Parameter 2'),
        ),
    ]
